/*
 *	base.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>

#include <sp/sp.h>
#include <sp/base.h>
#include <sp/memory.h>

#ifdef SP_USE_VECTOR_ENGINE
#include <sp/spPluginP.h>
#include <sp/spVectorPluginP.h>

extern spPlugin *sp_default_vector_plugin;
#endif

static int sp_randun_init = 0;

spBool spSrand(unsigned int seed)
{
#ifdef SP_USE_VECTOR_ENGINE
    if (sp_default_vector_plugin != NULL) {
        spVectorPluginInternalFuncList *flist;

        if ((flist = spGetVectorPluginInternalFuncList(sp_default_vector_plugin, 2, 0)) != NULL
             && flist->srand != NULL) {
             flist->srand(seed);
        }
        if ((flist = spGetVectorPluginInternalFuncList(sp_default_vector_plugin, 3, 0)) != NULL
             && flist->srand != NULL) {
             flist->srand(seed);
        }
    }
#endif
    
    srand(seed);
    sp_randun_init = 1;
    
    return SP_TRUE;
}

void spRandunNoInit(void)
{
    sp_randun_init = 1;
    return;
}

float spRandunf(void)
{
    float x;
    
    /* reset random function */
    if (!sp_randun_init) {
	srand((unsigned int)spGetProcessId());
	sp_randun_init = 1;
    }

    x = (float)rand() / ((float)RAND_MAX + 1.0f);

    return x;
}

float spRandun1f(void)
{
    return spRandunf();
}

double spRandun(void)
{
    double x;
    
    /* reset random function */
    if (!sp_randun_init) {
	srand((unsigned int)spGetProcessId());
	sp_randun_init = 1;
    }

    x = (double)rand() / ((double)RAND_MAX + 1.0);

    return x;
}

double spRandun1(void)
{
    return spRandun();
}

static int sp_gauss_use_rand = 0;

void spGaussUseRand(int flag)
{
    sp_gauss_use_rand = (flag ? 1 : 0);
    return;
}

/*
 *	mu : mean  sigma : standard deviation
 */
float spGaussf(float mu, float sigma)
{
    int i;
    float a, x;

    if (sp_gauss_use_rand) {
	for (i = 0, a = 0.0; i < 12; i++) {
	    a += (float)spRandun();
	}
    } else {
	for (i = 0, a = 0.0; i < 12; i++) {
	    a += (float)spRandun1();
	}
    }

    x = (a - 6.0f) * sigma + mu;

    return x;
}

double spGauss(double mu, double sigma)
{
    int i;
    double a, x;

    if (sp_gauss_use_rand) {
	for (i = 0, a = 0.0; i < 12; i++) {
	    a += spRandun();
	}
    } else {
	for (i = 0, a = 0.0; i < 12; i++) {
	    a += spRandun1();
	}
    }

    x = (a - 6.0) * sigma + mu;

    return x;
}

long spGcd(long x, long y)
{
    long a;

    while (y != 0) {
	a = x % y;
	x = y;
	y = a;
    }

    return x;
}

void spCExpf(float *xr, float *xi)
{
    float a;

    if (xr == NULL) {
	return;
    } else if (*xr == 0.0f) {
	*xr = cosf(*xi);
	*xi = sinf(*xi);
    } else if (xi != NULL && *xi != 0.0f) {
	a = expf(*xr);
	*xr = a * cosf(*xi);
	*xi = a * sinf(*xi);
    } else {
	*xr = expf(*xr);
    }

    return;
}

void spCExp(double *xr, double *xi)
{
    double a;

    if (xr == NULL) {
	return;
    } else if (*xr == 0.0) {
	*xr = cos(*xi);
	*xi = sin(*xi);
    } else if (xi != NULL && *xi != 0.0) {
	a = exp(*xr);
	*xr = a * cos(*xi);
	*xi = a * sin(*xi);
    } else {
	*xr = exp(*xr);
    }

    return;
}

void spCLogf(float *xr, float *xi)
{
    float a;

    if (xi != NULL && (*xr < 0.0f || *xi != 0.0f)) {
	a = (float)CABS(*xr, *xi);
	*xi = atan2f(*xi, *xr);
	*xr = logf(a);
    } else {
	if (*xr <= 0.0f) {
	    spwarning("warning: clogf: log of zero\n");
	    
	    *xr = logf((float)SP_TINY_NUMBER);
	} else {
	    *xr = logf(*xr);
	}
    }

    return;
}

void spCLog(double *xr, double *xi)
{
    double a;

    if (xi != NULL && (*xr < 0.0 || *xi != 0.0)) {
	a = CABS(*xr, *xi);
	*xi = atan2(*xi, *xr);
	*xr = log(a);
    } else {
	if (*xr <= 0.0) {
	    spwarning("warning: clog: log of zero\n");
	    
	    *xr = log(SP_TINY_NUMBER);
	} else {
	    *xr = log(*xr);
	}
    }

    return;
}

void spCLog10f(float *xr, float *xi)
{
    float a;

    if (xi != NULL && (*xr < 0.0f || *xi != 0.0f)) {
	a = (float)CABS(*xr, *xi);
	*xi = atan2f(*xi, *xr) / (float)SP_LN10;
	*xr = log10f(a);
    } else {
	if (*xr <= 0.0f) {
	    spwarning("warning: clog: log of zero\n");
	    
	    *xr = log10f((float)SP_TINY_NUMBER);
	} else {
	    *xr = log10f(*xr);
	}
    }

    return;
}

void spCLog10(double *xr, double *xi)
{
    double a;

    if (xi != NULL && (*xr < 0.0 || *xi != 0.0)) {
	a = CABS(*xr, *xi);
	*xi = atan2(*xi, *xr) / SP_LN10;
	*xr = log10(a);
    } else {
	if (*xr <= 0.0) {
	    spwarning("warning: clog: log of zero\n");
	    
	    *xr = log10(SP_TINY_NUMBER);
	} else {
	    *xr = log10(*xr);
	}
    }

    return;
}

typedef struct _spRandSortData
{
    int index;
    char *data;
    double randnum;
} spRandSortData;

#if defined(__STDC__) || defined(_WIN32) || defined(MACOS)
static int spRandSortNumCmp(const void *x, const void *y)
#else
static int spRandSortNumCmp(void *x, void *y)
#endif
{
    spRandSortData *xdata, *ydata;
    
    xdata = (spRandSortData *)x; ydata = (spRandSortData *)y;
    
    if (xdata->randnum < ydata->randnum) {
	return (-1);
    } else if (xdata->randnum > ydata->randnum) {
	return (1);
    } else {
	return (0);
    }
}

void spRandSort(void *data, int num, int size)
{
    int i;
    char *ptr;
    spRandSortData *data_list;

    data_list = xalloc(num, spRandSortData);

    ptr = data;
    for (i = 0; i < num; i++) {
	data_list[i].index = i;
	data_list[i].data = xalloc(size, char);
	memmove(data_list[i].data, ptr, size);
	
	data_list[i].randnum = randun1();
	spDebug(100, "spRandSort", "data_list[%d].randnum = %f\n", i, data_list[i].randnum);

	ptr = ptr + size;
    }
    
    qsort(data_list, (unsigned)num, (unsigned)sizeof(spRandSortData), spRandSortNumCmp);

    ptr = data;
    for (i = 0; i < num; i++) {
	memmove(ptr, data_list[i].data, size);
	xfree(data_list[i].data);
	
	ptr = ptr + size;
    }
    
    xfree(data_list);

    return;
}

void spDecibel(double *x, long length)
{
    long k;

    for (k = 0; k < length; k++) {
	x[k] = x[k] * x[k];
	if (x[k] <= 0.0) {
	    spwarning("warning: decibel: log of zero\n");
	    
	    x[k] = 10.0 * log10(SP_TINY_NUMBER);
	} else {
	    x[k] = 10.0 * log10(x[k]);
	}
    }

    return;
}

void spDecibelp(double *x, long length)
{
    long k;

    for (k = 0; k < length; k++) {
	if (x[k] <= 0.0) {
	    spwarning("warning: decibelp: log of zero\n");
	    
	    x[k] = 10.0 * log10(SP_TINY_NUMBER);
	} else {
	    x[k] = 10.0 * log10(x[k]);
	}
    }

    return;
}

